set ns [new Simulator]

set nf [open out.nam w]
$ns namtrace-all $nf

proc finish {} {
    global ns nf
    $ns flush-trace
    close $nf
    exec nam out.nam &
    exit 0
}

proc finishGraph {} {
    global ns nf fhands fnames ahands anames
    $ns flush-trace
    close $nf
    foreach fh $fhands {
	close $fh
    }
    exec ../xgraph.sh $fnames -geometry 800x400 &
    foreach aggf [array names ahands] {
	close $ahands($aggf)
    }
    exec ../xgraph.sh $anames -geometry 800x400 &
    exit 0
}

proc mkAggFile {cls} {
    global ahands anames
    set aname "agg$cls.tr"
    set fh [open $aname w]
    set ahands($cls) $fh
    lappend anames $aname
}

proc mkOutFile {} {
    global fhands fnames
    # Open and output file for data
    set onum [llength $fhands]
    set oname "out$onum.tr"
    set fh [open $oname w]
    lappend fhands $fh
    lappend fnames $oname
}

proc doCbrUdp { num source sink pktsize interval strt stp ramp } {
    global sinks aggs
    mkAggFile 2
    set aggs(2) [list]
    for {set i 0} {$i < $num} {set i [expr $i+1]} {
	set snk [mkCbrUdp $source $sink 2 $pktsize $interval green [expr $strt+$ramp*$i] $stp]
	lappend sinks $snk
	lappend aggs(2) $snk
    }
}

proc mkCbrUdp { source sink cls pktsize interval color strt stp } {
    # Get simulator instance
    set ns [Simulator instance]
    # Make UDP agent and attach to source
    set udp [new Agent/UDP]
    $ns attach-agent $source $udp
    # Create LossMonitor and attach to sink
    set loss [new Agent/LossMonitor]
    $ns attach-agent $sink $loss
    # Make CBR application
    set cbr [new Application/Traffic/CBR]
    $cbr set packetSize_ $pktsize
    $cbr set interval_ $interval
    # Attach to UDP agent
    $cbr attach-agent $udp
    # Connect source and sink
    $ns connect $udp $loss
    # Start it up
    $ns at $strt "$cbr start"
    $ns at $stp "$cbr stop"
    mkOutFile
    return $loss
}

proc doCbrTfrc { num source sink pktsize interval strt stp ramp } {
    global sinks aggs
    mkAggFile 1
    set aggs(1) [list]
    for {set i 0} {$i < $num} {set i [expr $i+1]} {
	set snk [mkCbrTfrc $source $sink 1 $pktsize $interval green [expr $strt+$ramp*$i] $stp]
	lappend sinks $snk
	lappend aggs(1) $snk
    }
}

proc mkCbrTfrc { source sink cls pktsize interval color strt stp } {
    # Get simulator instance
    set ns [Simulator instance]
    # Make TFRC agent and attach to source
    set tfrc [new Agent/TFRC]
    $tfrc set SndrType_ 1
    $tfrc set packetSize_ $pktsize
    $tfrc set maxqueue_ 2
    $ns attach-agent $source $tfrc
    # Create TFRCSink and attach to sink
    set loss [new Agent/TFRCSink]
    $ns attach-agent $sink $loss
    # Make CBR application
    set cbr [new Application/Traffic/CBR]
    $cbr set packetSize_ $pktsize
    $cbr set interval_ $interval
    # Attach to TFRC agent
    $cbr attach-agent $tfrc
    # Connect source and sink
    $ns connect $tfrc $loss
    # Start it up
    $ns at $strt "$cbr start"
    $ns at $stp "$cbr stop"
    mkOutFile
    return $loss
}

proc shortTcp { num source sink strt len ramp } {
    global sinks aggs
    mkAggFile 0
    for {set i 0} {$i < $num} {set i [expr $i+1]} {
	set start [expr $strt+$ramp*$i]
	set snk [mkTcp $source $sink 0 red $start [expr $start+$len]]
	lappend sinks $snk
	lappend aggs(0) $snk
    }
}

proc doTcp { num source sink strt stp ramp } {
    global sinks aggs
    mkAggFile 0
    set aggs(0) [list]
    for {set i 0} {$i < $num} {set i [expr $i+1]} {
	set snk [mkTcp $source $sink 0 red [expr $strt+$ramp*$i] $stp]
	lappend sinks $snk
	lappend aggs(0) $snk
    }
}

proc mkTcp { source sink cls color strt stp } {
    # Get simulator instance
    set ns [Simulator instance]
    # Create TCP flow
    set tcp [new Agent/TCP]
    $tcp set class_ $cls
    $ns color $cls $color
    set snk [new Agent/TCPSink]
    $ns attach-agent $source $tcp
    $ns attach-agent $sink $snk
    $ns connect $tcp $snk
    set ftp [new Application/FTP]
    $ftp attach-agent $tcp
    $ns at $strt "$ftp start"
    $ns at $stp "$ftp stop"
    mkOutFile
    return $snk
}

proc doTfrc { num source sink strt stp ramp } {
    global sinks aggs
    mkAggFile 3
    set aggs(3) [list]
    for {set i 0} {$i < $num} {set i [expr $i+1]} {
	set snk [mkTfrc $source $sink 0 red [expr $strt+$ramp*$i] $stp]
	lappend sinks $snk
	lappend aggs(3) $snk
    }
}

proc mkTfrc { source sink cls color strt stp } {
    # Get simulator instance
    set ns [Simulator instance]
    #Create TFRC flow
    set tfrc [new Agent/TFRC]
    $tfrc set class_ $cls
    $tfrc set maxXmtRate_ 1000000
    $tfrc set packetSize_ 1000
    $ns color $cls $color
    set snk [new Agent/TFRCSink]
    $ns attach-agent $source $tfrc
    $ns attach-agent $sink $snk
    $ns connect $tfrc $snk
    $ns at $strt "$tfrc start"
    $ns at $stp "$tfrc stop"
    mkOutFile
    return $snk
}

proc progress {} {
    set ns [Simulator instance]
    set now [$ns now]
    puts "$now ..."
    $ns at [expr $now+1.0] "progress"
}

proc getGraphData { time } {
    global fhands fnames sinks ahands aggs
    # Get simulator
    set ns [Simulator instance]
    # Get current time
    set now [$ns now]
    for {set i 0} {$i < [llength $sinks]} {incr i} {
	# Get bytes received
	set bw [[lindex $sinks $i] set bytes_]
	# Calculate bandwidth and write to file
	puts [lindex $fhands $i] "$now [expr $bw/$time*8]"
	# Reset byte count
	#[lindex $sinks $i] set bytes_ 0
    }
    # Get aggregate data
    foreach aggf [array names ahands] {
	set bw 0
	foreach snk $aggs($aggf) {
	    set bw [expr $bw + [$snk set bytes_]]
	    # Reset byte count
	    $snk set bytes_ 0
	}
	puts $ahands($aggf) "$now [expr $bw/$time*8]"
    }
    # Schedule next run
    $ns at [expr $now + $time] "getGraphData $time"
    return $now
}

# Basic topology
#
#   |----tcp flows----|
# [n0]---+       +---[n4]
#        |       |
#       [n2]----[n3]
#        |       |
# [n1]---+       +---[n5]
#   |---tfrc flows----|

# Create nodes
set n0 [$ns node]
set n1 [$ns node]
set n2 [$ns node]
set n3 [$ns node]
set n4 [$ns node]
set n5 [$ns node]

# TCP access links
$ns duplex-link $n0 $n2 10Mb 10ms DropTail
$ns duplex-link $n3 $n4 10Mb 10ms DropTail
# The chokepoint link (n2<-->n3)
$ns duplex-link $n2 $n3 1.5Mb 10ms DropTail
# TFRC access links
$ns duplex-link $n1 $n2 10Mb 10ms DropTail
$ns duplex-link $n3 $n5 10Mb 10ms DropTail

# Lists for output file manipulations
set fhands [list]
set fnames [list]
set sinks [list]
set anames [list]

#Create TCP flows
doTcp 4 $n0 $n4 2.0 19.0 0.2
#shortTcp 16 $n0 $n4 2.0 1.0 0.9

# Create TRFC flows
#doTfrc 1 $n1 $n5 1.0 20.0 0.2

# Create CBR/TFRC flows
# G.711, 10ms -- 96K bps
doCbrTfrc 1 $n1 $n5 120 0.010 1.0 20.0 0.1
# G.711, 20ms -- 80K bps
#doCbrTfrc 1 $n1 $n5 200 0.020 1.0 10.0 0.1
# G.711, 40ms -- 72K bps
#doCbrTfrc 1 $n1 $n5 360 0.040 1.0 10.0 0.1

# Create CBR/UDP flows
# G.711, 10ms -- 96K bps
#doCbrUdp 20 $n1 $n5 120 0.010 1.0 20.0 0.1

# Run simulation
$ns at 0.0 "getGraphData 0.100"
$ns at 1.0 "progress"
$ns at 20.0 "finishGraph"
$ns run